//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

#include "bodydynamics/factor/factor_force_torque_inertial.h"
#include "bodydynamics/processor/processor_force_torque_inertial.h"
#include "bodydynamics/sensor/sensor_force_torque_inertial.h"
#include "bodydynamics/internal/config.h"

#include <core/utils/params_server.h>
#include <core/yaml/parser_yaml.h>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>

#include <Eigen/Dense>

using namespace wolf;
using namespace bodydynamics::fti;

// Register StateBlock of type "L"
#include <core/state_block/state_block_derived.h>
#include <core/state_block/factory_state_block.h>
WOLF_REGISTER_STATEBLOCK_WITH_KEY(L, StateVector3d);

class Test_FactorForceTorqueInertialPreint : public testing::Test
{
  public:
    ProblemPtr                            problem;
    SensorForceTorqueInertialPtr          sensor;
    ProcessorForceTorqueInertialPtr processor;
    FrameBasePtr                          frame_origin, frame_last, frame;
    CaptureMotionPtr                      capture_origin, capture_last, capture;
    FeatureMotionPtr                      feature;
    FactorForceTorqueInertialPtr          factor;

    VectorXd data;
    MatrixXd data_cov;
    VectorXd delta;
    MatrixXd delta_cov;
    Vector6d bias;
    MatrixXd J_delta_bias;

  protected:
    void SetUp() override
    {
        data     = VectorXd::Random(12);  // [ a, w, f, t ]
        data_cov = MatrixXd::Identity(12, 12);

        std::string  wolf_root = _WOLF_BODYDYNAMICS_ROOT_DIR;
        ParserYaml   parser    = ParserYaml("problem_force_torque_inertial.yaml", wolf_root + "/test/yaml/");
        ParamsServer server    = ParamsServer(parser.getParams());
        problem                = Problem::autoSetup(server);

        sensor = std::static_pointer_cast<SensorForceTorqueInertial>(problem->getHardware()->getSensorList().front());
        processor = std::static_pointer_cast<ProcessorForceTorqueInertial>(sensor->getProcessorList().front());

        problem->print(4, 1, 1, 1);

        capture = std::make_shared<CaptureMotion>("CaptureForceTorqueInertial", 0.0, sensor, data, data_cov, nullptr);
        capture->process();

        problem->print(4, 1, 1, 1);

        capture_origin = std::static_pointer_cast<CaptureMotion>(processor->getOrigin());
        frame_origin   = capture_origin->getFrame();

        ASSERT_EQ(capture_origin->getTimeStamp(), frame_origin->getTimeStamp());
        ASSERT_EQ(capture_origin->getTimeStamp(), frame_origin->getCaptureOf(sensor)->getTimeStamp());

        capture = std::make_shared<CaptureMotion>("CaptureForceTorqueInertial", 0.1, sensor, data, data_cov, nullptr);
        capture->process();

        capture_last = std::static_pointer_cast<CaptureMotion>(processor->getLast());
        frame_last   = capture_last->getFrame();
    }
};

TEST(Dummy, force_load_libwolfbodydynamics_so_at_launch_time)
{
    ParamsSensorForceTorqueInertialPtr p = std::make_shared<ParamsSensorForceTorqueInertial>();
    p->acc_noise_std                     = 1;
    p->force_noise_std                   = 1;
    p->gyro_noise_std                    = 1;
    p->torque_noise_std                  = 1;
    SensorForceTorqueInertial s((Vector7d() << 1, 2, 3, 0, 0, 0, 1).finished(), p);
}

TEST_F(Test_FactorForceTorqueInertialPreint, constructor)
{
    feature = FeatureMotion::emplace<FeatureMotion>(capture_last, "FeatureFTI", VectorXd(19),
                                                    MatrixXd::Identity(18, 18), VectorXd(6), MatrixXd(18, 6));
    FactorForceTorqueInertial f(feature, capture_origin, processor, false);

    WOLF_INFO("f id: ", f.id());

    ASSERT_EQ(f.getCaptureOther(), capture_origin);
}

TEST_F(Test_FactorForceTorqueInertialPreint, emplace)
{
    feature = FeatureMotion::emplace<FeatureMotion>(capture_last, "FeatureFTI", VectorXd(19),
                                                    MatrixXd::Identity(18, 18), VectorXd(6), MatrixXd(18, 6));
    factor  = FactorBase::emplace<FactorForceTorqueInertial>(feature, feature, capture_origin, processor, false);
    // this ^^^ will create a warning "ADDING FACTOR id TO FEATURE id THAT IS NOT CONNECTED WITH PROBLEM."

    problem->print(4, 1, 1, 1); // feature and factor not in the tree since they belong to last_ptr with no KF.

    ASSERT_EQ(factor->getCaptureOther(), capture_origin);
    ASSERT_EQ(factor->getCapture(), capture_last);
}

int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);

    return RUN_ALL_TESTS();
}
