//--------LICENSE_START--------
//
// Copyright (C) 2020,2021,2022 Institut de Robòtica i Informàtica Industrial, CSIC-UPC.
// Authors: Joan Solà Ortega (jsola@iri.upc.edu)
// All rights reserved.
//
// This file is part of WOLF
// WOLF is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
//--------LICENSE_END--------

#include "bodydynamics/factor/factor_force_torque_inertial_dynamics.h"
#include "bodydynamics/processor/processor_force_torque_inertial_dynamics.h"
#include "bodydynamics/sensor/sensor_force_torque_inertial.h"
#include "bodydynamics/internal/config.h"
#include "bodydynamics/capture/capture_force_torque_inertial.h"

#include <core/sensor/factory_sensor.h>

#include <core/utils/utils_gtest.h>
#include <core/utils/logging.h>
#include <core/utils/params_server.h>
#include <core/yaml/parser_yaml.h>
#include <core/state_block/state_block_derived.h>
#include <core/state_block/factory_state_block.h>
#include <core/ceres_wrapper/solver_ceres.h>
#include <core/math/rotations.h>
#include <core/tree_manager/tree_manager_sliding_window.h>

#include <Eigen/Dense>
#include <Eigen/Geometry>

using namespace wolf;
using namespace bodydynamics::fti;

WOLF_REGISTER_STATEBLOCK_WITH_KEY(L, StateVector3d);

class Test_SolveProblemForceTorqueInertialDynamics_yaml : public testing::Test
{
  public:
    ProblemPtr                              P;
    SensorForceTorqueInertialPtr            S;
    ProcessorForceTorqueInertialDynamicsPtr p;
    SolverCeresPtr                          solver;
    Vector6d                                bias_true;
    Vector3d                                cdm_true;
    Vector3d                                inertia_true;
    double                                  mass_true;

  protected:
    void SetUp() override
    {
        std::string wolf_root = _WOLF_BODYDYNAMICS_ROOT_DIR;
        ParserYaml  parser =
            ParserYaml("problem_force_torque_inertial_dynamics_solve_test.yaml", wolf_root + "/test/yaml/");
        ParamsServer server = ParamsServer(parser.getParams());
        P                   = Problem::autoSetup(server);
        // CERES WRAPPER
        solver = std::make_shared<SolverCeres>(P);
        // solver->getSolverOptions().max_num_iterations               = 1e4;
        // solver->getSolverOptions().function_tolerance               = 1e-15;
        // solver->getSolverOptions().gradient_tolerance               = 1e-15;
        // solver->getSolverOptions().parameter_tolerance              = 1e-15;

        S = std::static_pointer_cast<SensorForceTorqueInertial>(P->getHardware()->getSensorList().front());
        p = std::static_pointer_cast<ProcessorForceTorqueInertialDynamics>(S->getProcessorList().front());

        bias_true    << 0,0,0, 0,0,0;
        cdm_true     = {0, 0, 0.0341};
        inertia_true = {0.017598, 0.017957, 0.029599};
        mass_true    = 1.952;
    }
};

TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, force_registraion)
{
    FactorySensor::registerCreator("SensorForceTorqueInertial", SensorForceTorqueInertial::create);
}

// Hover test.
TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, mass_only_hovering)
{
    double mass_guess = S->getStateBlock('m')->getState()(0);

    // Data
    VectorXd data      = VectorXd::Zero(12);      // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();              // accelerometer
    data.segment<3>(6) = -mass_true * gravity();  // force
    MatrixXd data_cov  = MatrixXd::Identity(12, 12) * 1e-6;

    // We fix the parameters of the sensor except for the mass
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->fix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->unfix();
    // S->setStateBlockDynamic('I');

    CaptureMotionPtr C0_0 = std::make_shared<CaptureForceTorqueInertial>(0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureForceTorqueInertial>(0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureForceTorqueInertial>(0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureForceTorqueInertial>(0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureForceTorqueInertial>(0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureForceTorqueInertial>(1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->fix();
    C_KF1->getFrame()->fix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);

    WOLF_INFO("True mass     : ", mass_true, " Kg.");
    WOLF_INFO("Guess mass    : ", mass_guess, " Kg.");
    WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
    WOLF_INFO("-----------------------------");

    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0; t <= 7.0; t += 1.0)
    {
        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data, data_cov, nullptr);
        C->process();
        p->getOrigin()->getFrame()->fix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
        WOLF_INFO("-----------------------------");
    }

    P->print(4, 1, 1, 1);

    ASSERT_NEAR(S->getStateBlock('m')->getState()(0), mass_true, 1e-3);
}

TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, cdm_only_hovering)
{
    Vector3d cdm_guess (0.01, 0.01, 0.033);
    S->getStateBlock('C')->setState(cdm_guess);
    S->getStateBlock('m')->setState(Vector1d(mass_true));
    

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data.segment<3>(6) = -mass_true * gravity();
    MatrixXd data_cov  = MatrixXd::Identity(12, 12);

    // We fix the parameters of the sensor except for the cdm
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->fix();
    // S->setStateBlockDynamic('I');

    CaptureMotionPtr C0_0 = std::make_shared<CaptureForceTorqueInertial>(0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureForceTorqueInertial>(0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureForceTorqueInertial>(0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureForceTorqueInertial>(0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureForceTorqueInertial>(0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureForceTorqueInertial>(1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->fix();
    C_KF1->getFrame()->getStateBlock('L')->setState(Vector3d::Zero());  // Make sure L is not moving due to torques
    C_KF1->getFrame()->fix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);  // should show a very low iteration number (possibly 1)

    WOLF_INFO("True cdm     : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess cdm    : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated cdm: ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("-----------------------------");

    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0; t <= 22.0; t += 1.0)
    {
        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data, data_cov, nullptr);
        C->process();
        p->getOrigin()->getFrame()->getStateBlock('L')->setState(
            Vector3d::Zero());  // Make sure L is not moving due to torques
        p->getOrigin()->getFrame()->fix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
        WOLF_INFO("-----------------------------");
    }

    ASSERT_MATRIX_APPROX(
        S->getStateBlock('C')->getState().head(2), cdm_true.head(2), 1e-5);  // cdm_z is not observable while hovering
}

TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, mass_and_cdm_hovering)
{
    Vector3d cdm_guess  (0.005, 0.005, 0.05);
    double   mass_guess = 1.9;
    S->getStateBlock('C')->setState(cdm_guess);
    S->getStateBlock('m')->setState(Vector1d(mass_guess));

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data.segment<3>(6) = -mass_true * gravity();
    MatrixXd data_cov  = MatrixXd::Identity(12, 12);

    // We fix the parameters of the sensor except for the cdm and the mass
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->unfix();
    // S->setStateBlockDynamic('I');

    CaptureMotionPtr C0_0 = std::make_shared<CaptureForceTorqueInertial>(0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureForceTorqueInertial>(0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureForceTorqueInertial>(0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureForceTorqueInertial>(0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureForceTorqueInertial>(0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureForceTorqueInertial>(1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->fix();
    C_KF1->getFrame()->getStateBlock('L')->setState(Vector3d::Zero());  // Make sure L is not moving due to torques
    C_KF1->getFrame()->fix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);  // should show a very low iteration number (possibly 1)

    WOLF_INFO("True mass     : ", mass_true, " Kg.");
    WOLF_INFO("True cdm      : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess mass    : ", mass_guess, " Kg.");
    WOLF_INFO("Guess cdm     : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
    WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("-----------------------------");

    // Do a number of iterations with more keyframes, solving so that the calibration gets better and better
    // Here we take advantage of the sliding window, thereby getting rid progressively
    // of the old factors, which contained calibration parameters far from the converged values,
    // which if not eliminated contaminate the overall solution.
    // This is due to these first factors relying on the linearization Jacobian to correct the
    // poorly preintegrated delta.
    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0; t <= 50.0; t += 1.0)
    {
        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data, data_cov, nullptr);
        C->setTimeStamp(t);
        C->process();
        p->getOrigin()->getFrame()->getStateBlock('L')->setState(
            Vector3d::Zero());  // Make sure L is not moving due to torques
        p->getOrigin()->getFrame()->fix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
        WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
        WOLF_INFO("-----------------------------");
    }

    auto cdm_estimated  = S->getStateBlock('C')->getState();
    auto mass_estimated = S->getStateBlock('m')->getState()(0);

    ASSERT_NEAR(mass_estimated, mass_true, 1e-6);
    ASSERT_MATRIX_APPROX(cdm_estimated.head(2), cdm_true.head(2), 1e-6);  // cdm_z is not observable while hovering
}

TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, cdm_and_ang_mom_hovering)
{
    Vector3d cdm_guess (0.02, 0.02, 0.033);
    S->getStateBlock('C')->setState(cdm_guess);
    S->getStateBlock('m')->setState(Vector1d(mass_true));

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data.segment<3>(6) = -mass_true * gravity();
    MatrixXd data_cov  = MatrixXd::Identity(12, 12);

    // We fix the parameters of the sensor except for the cdm
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->fix();
    // S->setStateBlockDynamic('I');

    CaptureMotionPtr C0_0 = std::make_shared<CaptureForceTorqueInertial>(0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureForceTorqueInertial>(0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureForceTorqueInertial>(0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureForceTorqueInertial>(0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureForceTorqueInertial>(0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureForceTorqueInertial>(1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->fix();
    C_KF0->getFrame()->getStateBlock('L')->unfix();
    C_KF1->getFrame()->fix();
    C_KF1->getFrame()->getStateBlock('L')->unfix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);  // should show a very low iteration number (possibly 1)

    WOLF_INFO("True cdm     : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess cdm    : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated cdm: ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("-----------------------------");

    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0; t <= 30.0; t += 1.0)
    {
        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data, data_cov, nullptr);
        C->process();
        p->getOrigin()->getFrame()->fix();
        p->getOrigin()->getFrame()->getStateBlock('L')->unfix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
        WOLF_INFO("Estimated ang mom : ",
                  p->getOrigin()->getFrame()->getStateBlock('L')->getState().transpose(),
                  " m^2 kg/s.");
        WOLF_INFO("-----------------------------");
    }

    ASSERT_MATRIX_APPROX(
        S->getStateBlock('C')->getState().head(2), cdm_true.head(2), 1e-5);  // cdm_z is not observable while hovering
}

TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, mass_and_cdm_and_ang_mom_hovering)
{
    Vector3d cdm_guess  (0.05, 0.05, 0.05);
    double   mass_guess = 1.50;
    S->getStateBlock('C')->setState(cdm_guess);
    S->getStateBlock('m')->setState(Vector1d(mass_guess));

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data.segment<3>(6) = -mass_true * gravity();
    MatrixXd data_cov  = MatrixXd::Identity(12, 12);

    // We fix the parameters of the sensor except for the cdm and the mass
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->unfix();
    // S->setStateBlockDynamic('I');

    CaptureMotionPtr C0_0 = std::make_shared<CaptureForceTorqueInertial>(0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureForceTorqueInertial>(0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureForceTorqueInertial>(0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureForceTorqueInertial>(0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureForceTorqueInertial>(0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureForceTorqueInertial>(1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->fix();
    C_KF0->getFrame()->getStateBlock('L')->unfix();
    C_KF1->getFrame()->fix();
    C_KF1->getFrame()->getStateBlock('L')->unfix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);  // should show a very low iteration number (possibly 1)

    WOLF_INFO("True mass     : ", mass_true, " Kg.");
    WOLF_INFO("True cdm      : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess mass    : ", mass_guess, " Kg.");
    WOLF_INFO("Guess cdm     : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
    WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("-----------------------------");

    // Do a number of iterations with more keyframes, solving so that the calibration gets better and better
    // Here we take advantage of the sliding window, thereby getting rid progressively
    // of the old factors, which contained calibration parameters far from the converged values,
    // which if not eliminated contaminate the overall solution.
    // This is due to these first factors relying on the linearization Jacobian to correct the
    // poorly preintegrated delta.
    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0; t <= 26.0; t += 1.0)
    {
        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data, data_cov, nullptr);
        C->setTimeStamp(t);
        C->process();
        p->getOrigin()->getFrame()->fix();
        p->getOrigin()->getFrame()->getStateBlock('L')->unfix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
        WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
        WOLF_INFO("Estimated ang mom : ",
                  p->getOrigin()->getFrame()->getStateBlock('L')->getState().transpose(),
                  " m^2 kg/s.");

        WOLF_INFO("-----------------------------");
    }

    auto cdm_estimated  = S->getStateBlock('C')->getState();
    auto mass_estimated = S->getStateBlock('m')->getState()(0);

    ASSERT_NEAR(mass_estimated, mass_true, 1e-6);
    ASSERT_MATRIX_APPROX(cdm_estimated.head(2), cdm_true.head(2), 1e-6);  // cdm_z is not observable while hovering
}

// Here we only fix P and O from each key frame
TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, hovering_test_without_fixing)
{
    Vector3d cdm_guess (0.05, 0.05, 0.05);
    double   mass_guess = 1.50;
    S->getStateBlock('C')->setState(cdm_guess);
    S->getStateBlock('m')->setState(Vector1d(mass_guess));

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data.segment<3>(6) = -mass_true * gravity();
    MatrixXd data_cov  = MatrixXd::Identity(12, 12);

    // We fix the parameters of the sensor except for the cdm and the mass
    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->unfix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->fix();
    S->getStateBlock('m')->unfix();
    // S->setStateBlockDynamic('I');

    CaptureMotionPtr C0_0 = std::make_shared<CaptureForceTorqueInertial>(0.0, S, data, data_cov, nullptr);
    CaptureMotionPtr C1_0 = std::make_shared<CaptureForceTorqueInertial>(0.2, S, data, data_cov, nullptr);
    CaptureMotionPtr C2_0 = std::make_shared<CaptureForceTorqueInertial>(0.4, S, data, data_cov, nullptr);
    CaptureMotionPtr C3_0 = std::make_shared<CaptureForceTorqueInertial>(0.6, S, data, data_cov, nullptr);
    CaptureMotionPtr C4_0 = std::make_shared<CaptureForceTorqueInertial>(0.8, S, data, data_cov, nullptr);
    CaptureMotionPtr C5_1 = std::make_shared<CaptureForceTorqueInertial>(1.0, S, data, data_cov, nullptr);

    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());
    C1_0->process();
    C2_0->process();
    C3_0->process();
    C4_0->process();
    C5_1->process();

    CaptureMotionPtr C_KF1 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->unfix();
    C_KF0->getFrame()->getStateBlock('P')->fix();
    C_KF0->getFrame()->getStateBlock('O')->fix();
    C_KF1->getFrame()->unfix();
    C_KF1->getFrame()->getStateBlock('P')->fix();
    C_KF1->getFrame()->getStateBlock('O')->fix();

    P->print(4, 1, 1, 1);

    WOLF_INFO("======== SOLVE PROBLEM =======")
    std::string report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
    WOLF_INFO(report);  // should show a very low iteration number (possibly 1)

    WOLF_INFO("True mass     : ", mass_true, " Kg.");
    WOLF_INFO("True cdm      : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess mass    : ", mass_guess, " Kg.");
    WOLF_INFO("Guess cdm     : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
    WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("-----------------------------");

    // Do a number of iterations with more keyframes, solving so that the calibration gets better and better
    // Here we take advantage of the sliding window, thereby getting rid progressively
    // of the old factors, which contained calibration parameters far from the converged values,
    // which if not eliminated contaminate the overall solution.
    // This is due to these first factors relying on the linearization Jacobian to correct the
    // poorly preintegrated delta.
    for (TimeStamp t = p->getLast()->getTimeStamp() + 1.0; t <= 30.0; t += 1.0)
    {
        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data, data_cov, nullptr);
        C->setTimeStamp(t);
        C->process();  // En aquesta línia el programa s'aborta
        p->getOrigin()->getFrame()->unfix();
        p->getOrigin()->getFrame()->getStateBlock('P')->fix();
        p->getOrigin()->getFrame()->getStateBlock('O')->fix();
        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated mass: ", S->getStateBlock('m')->getState(), " Kg.");
        WOLF_INFO("Estimated cdm : ", S->getStateBlock('C')->getState().transpose(), " m.");
        WOLF_INFO("Estimated ang mom : ",
                  p->getOrigin()->getFrame()->getStateBlock('L')->getState().transpose(),
                  " m^2 kg/s.");

        WOLF_INFO("-----------------------------");
    }

    auto cdm_estimated  = S->getStateBlock('C')->getState();
    auto mass_estimated = S->getStateBlock('m')->getState()(0);

    ASSERT_NEAR(mass_estimated, mass_true, 1e-6);
    ASSERT_MATRIX_APPROX(cdm_estimated.head(2), cdm_true.head(2), 1e-6);  // cdm_z is not observable while hovering
}

// Here we only fix P and O from each key frame
TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, rotation_around_z_axis_test)
{
    Vector3d inertia_guess (0.01, 0.01, 0.02);
    S->getStateBlock('i')->setState(inertia_guess);

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    Vector3d ang_vel_measure_initial(0, 0, 0);
    data.segment<3>(3) = ang_vel_measure_initial;
    data.segment<3>(6) = -mass_true * gravity();
    Vector3d torque_measure_constant(0, 0, 1);
    data.segment<3>(9) = torque_measure_constant;
    MatrixXd data_cov  = MatrixXd::Identity(12, 12);

    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->fix();
    S->getStateBlock('C')->fix();
    S->getStateBlock('i')->unfix();
    S->getStateBlock('m')->fix();
    // S->setStateBlockDynamic('I');

    CaptureMotionPtr C0_0 = std::make_shared<CaptureForceTorqueInertial>(0.0, S, data, data_cov, nullptr);
    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->unfix();
    C_KF0->getFrame()->getStateBlock('P')->fix();
    C_KF0->getFrame()->getStateBlock('P')->setState(Vector3d(0, 0, 0));
    C_KF0->getFrame()->getStateBlock('O')->fix();
    C_KF0->getFrame()->getStateBlock('O')->setState((Quaterniond(1, 0, 0, 0)).coeffs());

    // initialiation of all the variables needed in the following iteration process
    Vector3d    torque_true(0, 0, 1);
    Vector3d    dL;
    Vector3d    L_true(0, 0, 0);
    Vector3d    ang_vel_true(0, 0, 0);
    Vector3d    ang_acc_true;
    Vector3d    angle_true;
    Quaterniond quat_true(1, 0, 0, 0);
    double      dt = 1;  // time increment
    Vector3d    position_data(0, 0, 0);
    Quaterniond q_data;
    std::string report;

    for (TimeStamp t = 1.0; t <= 18.0; t += 1.0)
    {
        // SIMULATOR

        // calculate dL = torque*dt
        dL = torque_true * dt;
        // actualize L = L + dt
        L_true = L_true + dL;  // L = L + dt
        // w = L*i_inv  we actualize the data
        ang_vel_true(0)    = L_true(0) / inertia_true(0);
        ang_vel_true(1)    = L_true(1) / inertia_true(1);
        ang_vel_true(2)    = L_true(2) / inertia_true(2);
        data.segment<3>(3) = ang_vel_true;

        ang_acc_true(0) = torque_true(0) / inertia_true(0);
        ang_acc_true(1) = torque_true(1) / inertia_true(1);
        ang_acc_true(2) = torque_true(2) / inertia_true(2);

        angle_true = ang_vel_true * dt + 0.5 * ang_acc_true * dt * dt;
        quat_true     = quat_true * exp_q(angle_true);

        // ESTIMATOR

        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data, data_cov, nullptr);
        C->setTimeStamp(t);
        C->process();

        p->getOrigin()->getFrame()->unfix();
        p->getOrigin()->getFrame()->getStateBlock('P')->fix();
        p->getOrigin()->getFrame()->getStateBlock('P')->setState(position_data);
        q_data = quat_true;
        p->getOrigin()->getFrame()->getStateBlock('O')->fix();
        p->getOrigin()->getFrame()->getStateBlock('O')->setState(q_data.coeffs());

        report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);
        WOLF_INFO("Estimated inertia: ", S->getStateBlock('i')->getState().transpose(), " m^2 Kg.");
        WOLF_INFO("-----------------------------");
    }

    WOLF_INFO("True inertia     : ", inertia_true.transpose(), " m^2 Kg.");
    WOLF_INFO("Guess inertia    : ", inertia_guess.transpose(), " m^2 Kg.");
    WOLF_INFO("Estimated inertia: ", S->getStateBlock('i')->getState().transpose(), " m^2 Kg.");
    WOLF_INFO("-----------------------------");
    auto inertia_estimated = S->getStateBlock('i')->getState();

    ASSERT_NEAR(inertia_estimated(2), inertia_true(2), 1e-6);
}

TEST_F(Test_SolveProblemForceTorqueInertialDynamics_yaml, rotation_test_unfixing)
{
    Vector6d bias_guess; bias_guess << 0,0,0, 0,0,0;
    Vector3d cdm_guess (0.005, 0.005, 0.1);
    S->getStateBlock('C')->setState(cdm_guess);
    Vector3d inertia_guess (0.01, 0.01, 0.02);
    S->getStateBlock('i')->setState(inertia_guess);
    double mass_guess = 2.0;
    S->getStateBlock('m')->setState(Vector1d(mass_guess));

    Vector3d    acc_true = -gravity();
    Vector3d    ang_vel_true (0, 0, 0);
    Vector3d    force_true = -mass_true * gravity();
    Vector3d    torque_true (0, 0, 0.01);
    Vector3d    L_true (0, 0, 0);
    Vector3d    position_true (0, 0, 0);
    Quaterniond quat_true (1, 0, 0, 0);
    Vector3d    ang_acc_true;
    ang_acc_true(0) = torque_true(0) / inertia_true(0);
    ang_acc_true(1) = torque_true(1) / inertia_true(1);
    ang_acc_true(2) = torque_true(2) / inertia_true(2);
    Vector3d    angle_true(0,0,0);

    Vector3d    position_data = position_true;
    Vector4d    orient_data = quat_true.coeffs();

    // Data
    VectorXd data      = VectorXd::Zero(12);  // [ a, w, f, t ]
    data.segment<3>(0) = -gravity();
    data.segment<3>(3) = ang_vel_true;
    data.segment<3>(6) = -mass_true * gravity();
    data.segment<3>(9) = torque_true;
    MatrixXd data_cov  = 1 * MatrixXd::Identity(12, 12);

    S->getStateBlock('P')->fix();
    S->getStateBlock('O')->fix();
    S->getStateBlock('I')->unfix();
    S->getStateBlock('C')->unfix();
    S->getStateBlock('i')->unfix();
    S->getStateBlock('m')->unfix();
    S->setStateBlockDynamic('I', false);

    CaptureMotionPtr C0_0 = std::make_shared<CaptureForceTorqueInertial>(0.0, S, data, data_cov, nullptr);
    C0_0->process();
    CaptureMotionPtr C_KF0 = std::static_pointer_cast<CaptureMotion>(p->getOrigin());

    C_KF0->getFrame()->unfix();
    C_KF0->getFrame()->getStateBlock('P')->fix();
    C_KF0->getFrame()->getStateBlock('P')->setState(position_data);
    C_KF0->getFrame()->getStateBlock('O')->fix();
    C_KF0->getFrame()->getStateBlock('O')->setState(orient_data);

    // initialiation of all the variables needed in the following iteration process
    
    Vector3d    dL;
    double      dt = 0.1;  // time increment
    std::string report;
    unsigned int last_kf_id = p->getOrigin()->getFrame()->id();

    for (TimeStamp t = dt; t <= 8.0; t += dt)
    {
        // SIMULATOR

        // angular momentum
        L_true         += torque_true * dt;  // FIXME JS: it seems this is not needed!

        // ang velocity
        ang_vel_true   += ang_acc_true * dt;

        // orientation (quaternion and total angle) 
        // FIXME JS: this should be on top but it works better here, weird...
        quat_true      *= exp_q(ang_vel_true * dt /*+ 0.5 * ang_acc_true * dt * dt*/);
        angle_true     += ang_vel_true * dt /*+ 0.5 * ang_acc_true * dt * dt*/;
        
        // // w = L*i_inv  we actualize the data
        // ang_vel_true(0)    = L_true(0) / inertia_true(0);
        // ang_vel_true(1)    = L_true(1) / inertia_true(1);
        // ang_vel_true(2)    = L_true(2) / inertia_true(2);

        // FTI measurements
        data.segment<3>(0) = acc_true;
        data.segment<3>(3) = ang_vel_true;
        data.segment<3>(6) = force_true;
        data.segment<3>(9) = torque_true;

        // Pose measurements (simulate motion capture)
        position_data   = position_true;
        orient_data     = quat_true.coeffs();


        // ESTIMATOR

        CaptureMotionPtr C = std::make_shared<CaptureForceTorqueInertial>(t, S, data, data_cov, nullptr);
        C->process();

        // check if new KF
        if (last_kf_id != p->getOrigin()->getFrame()->id())
        {
            last_kf_id = p->getOrigin()->getFrame()->id();       
            
            // fix measured position and orientation
            p->getOrigin()->getFrame()->unfix();
            p->getOrigin()->getFrame()->getStateBlock('P')->fix();
            p->getOrigin()->getFrame()->getStateBlock('P')->setState(position_data);
            p->getOrigin()->getFrame()->getStateBlock('O')->fix();
            p->getOrigin()->getFrame()->getStateBlock('O')->setState(orient_data);

            report = solver->solve(wolf::SolverManager::ReportVerbosity::FULL);

            WOLF_INFO("Total angle turned      : ", angle_true.transpose(), " rad.");
            WOLF_INFO("Estimated IMU bias      : ", S->getStateBlock('I')->getState().transpose(), " m/s2 | rad/s.");
            WOLF_INFO("Estimated center of mass: ", S->getStateBlock('C')->getState().transpose(), " m.");
            WOLF_INFO("Estimated inertia       : ", S->getStateBlock('i')->getState().transpose(), " m^2 Kg.");
            WOLF_INFO("Estimated mass          : ", S->getStateBlock('m')->getState()(0), " Kg.");
            WOLF_INFO("-----------------------------");
        }
    }

    WOLF_INFO("True IMU bias           : ", bias_true.transpose(), " m/s2 | rad/s.");
    WOLF_INFO("Guess IMU bias          : ", bias_guess.transpose(), " m/s2 | rad/s.");
    WOLF_INFO("Estimated IMU bias      : ", S->getStateBlock('I')->getState().transpose(), " m/s2 | rad/s.");
    WOLF_INFO("True center of mass     : ", cdm_true.transpose(), " m.");
    WOLF_INFO("Guess center of mass    : ", cdm_guess.transpose(), " m.");
    WOLF_INFO("Estimated center of mass: ", S->getStateBlock('C')->getState().transpose(), " m.");
    WOLF_INFO("True inertia            : ", inertia_true.transpose(), " m^2 Kg.");
    WOLF_INFO("Guess inertia           : ", inertia_guess.transpose(), " m^2 Kg.");
    WOLF_INFO("Estimated inertia       : ", S->getStateBlock('i')->getState().transpose(), " m^2 Kg.");
    WOLF_INFO("True mass               : ", mass_true, " Kg.");
    WOLF_INFO("Guess mass              : ", mass_guess, " Kg.");
    WOLF_INFO("Estimated mass          : ", S->getStateBlock('m')->getState()(0), " Kg.");
    WOLF_INFO("-----------------------------");

    auto bias_estimated    = S->getStateBlock('I')->getState();
    auto cdm_estimated     = S->getStateBlock('C')->getState();
    auto inertia_estimated = S->getStateBlock('i')->getState();
    auto mass_estimated    = S->getStateBlock('m')->getState()(0);

    ASSERT_MATRIX_APPROX(bias_estimated, bias_true, 1e-4);
    ASSERT_MATRIX_APPROX(cdm_estimated.head(2), cdm_true.head(2), 1e-6);  // cdm_z is not observable while hovering
    ASSERT_NEAR(inertia_estimated(2), inertia_true(2), 1e-6);             // Ix and Iy not observable in this test
    ASSERT_NEAR(mass_estimated, mass_true, 1e-5);
}

int main(int argc, char **argv)
{
    testing::InitGoogleTest(&argc, argv);
    // ::testing::GTEST_FLAG(filter) = "Test_SolveProblemForceTorqueInertialDynamics_yaml.rotation_test_unfixing";
    return RUN_ALL_TESTS();
}
